/*
 *  Copyright (C) 2003-2004 Curt Coder
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string>
#include <SDL.h>
#include "agipic.h"
#include "gfx.h"
#include "font.h"

static SYSTEM sys;

// Screen

void InitScreen()
{
	int w = sys.w;
	int h = sys.h;

	if (sys.fDblSize)
	{
		w *= 2;
		h *= 2;
	}

	sys.screen = SDL_SetVideoMode(w, h, 32, SDL_HWSURFACE | SDL_DOUBLEBUF);
}

void SetWindowTitle(char *szTitle)
{
	SDL_WM_SetCaption(szTitle, NULL);
}

void SetScreenPar(int w, int h, char* fontdata)
{
	memcpy(sys.font, fontdata, sizeof(sys.font));

	sys.w = w;
	sys.h = h;
	sys.maxCol = sys.w / 8;
	sys.maxRow = sys.h / 8;
	sys.attr = IDA_DEFAULT;
	InitScreen();
}

void UpdateScreen()
{
	SDL_Flip(sys.screen);
}

void ClearScreen(int attr)
{
	int forecolor;

	sys.attr = attr;
	forecolor = (sys.attr & 0xF0) / 0x10;

	SDL_FillRect(sys.screen, NULL, EGACOLOR(forecolor));
}

void ClearGfxScreen(int color)
{
	SDL_Rect dest;
	dest.x = 0;
	dest.y = 0;
	dest.w = sys.screen->w;
	dest.h = IDI_MAX_ROW_PIC * GetFontHeight();

	LockScreen();
	SDL_FillRect(sys.screen, &dest, EGACOLOR(color));
	UnlockScreen();
}

void ClearRow(int row)
{
	for (int col = 0; col < sys.maxCol; col++) DrawStrAttr(row, col, IDA_DEFAULT, " ");
}

void ClearTextArea()
{
	for (int row = IDI_MAX_ROW_PIC; row < sys.maxRow; row++) ClearRow(row);
}

// Fonts

int GetFontWidth() 
{
	return sys.fDblSize ? 16 : 8;
}

int GetFontHeight() 
{
	return GetFontWidth();
}

// Text drawing routines

void DrawStrAttr(int row, int col, int attr, char *buffer)
{
	int code;

	LockScreen();

	for (int iChar = 0; iChar < (int)strlen(buffer); iChar++)
	{
		code = buffer[iChar];

		switch (code)
		{
		case '\n':
		case 0x8D:
			if (++row == sys.maxRow) return;
			col = 0;
			break;

		case '|':
			// swap attribute nibbles
			break;

		default:
			DrawChar(col * 8, row * 8, attr, code, (char*)sys.font);

			if (++col == sys.maxCol)
			{
				col = 0;
				if (++row == sys.maxRow) return;
			}
		}
	}

	UnlockScreen();
}

void DrawStr(int row, int col, char *str)
{
	DrawStrAttr(row, col, sys.attr, str);
}

void DrawStrAttrMiddle(int row, int attr, char *str)
{
	int col = (sys.maxCol / 2) - (strlen(str) / 2);
	DrawStrAttr(row, col, attr, str);
}

void DrawStrAttrRight(int row, int attr, char *str)
{
	int col = (sys.maxCol - strlen(str));
	DrawStrAttr(row, col, attr, str);
}

// Low-Level Graphics Routines

void _PutPixel(SDL_Surface *surface, int x, int y, Uint32 color)
{
	int bpp = surface->format->BytesPerPixel;
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        *p = color;
        break;
    case 2:
        *(Uint16 *)p = color;
        break;
    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
            p[0] = (color >> 16) & 0xff;
            p[1] = (color >> 8) & 0xff;
            p[2] = color & 0xff;
        } else {
            p[0] = color & 0xff;
            p[1] = (color >> 8) & 0xff;
            p[2] = (color >> 16) & 0xff;
        }
        break;
    case 4:
		*(Uint32 *)p = color;
		break;
    }
}

Uint32 _GetPixel(SDL_Surface *Surface, Sint32 X, Sint32 Y)
{
	Uint8  *bits;
	Uint32 Bpp;

	if (X<0) puts("x too small in GetPixel!");
	if (X>=Surface->w) puts("x too big in GetPixel!");

	Bpp = Surface->format->BytesPerPixel;

	bits = ((Uint8 *)Surface->pixels)+Y*Surface->pitch+X*Bpp;

	// Get the pixel
	switch(Bpp) {
	case 1:
		return *((Uint8 *)Surface->pixels + Y * Surface->pitch + X);
		break;
	case 2:
		return *((Uint16 *)Surface->pixels + Y * Surface->pitch/2 + X);
		break;
	case 3: { // Format/endian independent
		Uint8 r, g, b;
		r = *((bits)+Surface->format->Rshift/8);
		g = *((bits)+Surface->format->Gshift/8);
		b = *((bits)+Surface->format->Bshift/8);
		return SDL_MapRGB(Surface->format, r, g, b);
		}
		break;
	case 4:
		return *((Uint32 *)Surface->pixels + Y * Surface->pitch/4 + X);
		break;
	}

	return -1;
}

void _LockScreen(SDL_Surface *surface)
{
	if (SDL_MUSTLOCK(surface)) if (SDL_LockSurface(surface) < 0) return;
}

void _UnlockScreen(SDL_Surface *surface)
{
	if (SDL_MUSTLOCK(surface)) SDL_UnlockSurface(surface);
}

// Graphics wrapper routines

void PutPixel(int x, int y, Uint32 color)
{
	if (sys.fDblSize)
	{
		_PutPixel(sys.screen,	(x * 2),		(y * 2),		color);
		_PutPixel(sys.screen,	(x * 2) + 1,	(y * 2),		color);
		_PutPixel(sys.screen,	(x * 2),		(y * 2) + 1,	color);
		_PutPixel(sys.screen,	(x * 2) + 1,	(y * 2) + 1,	color);
	}
	else
	{
		_PutPixel(sys.screen, x, y, color);
	}
}

void PutPixelAGI(int x, int y, int color)
{
	Uint32 colOdd = CGACOLORODD(color);
	Uint32 colEven = CGACOLOREVEN(color);

	int vx = x * 2;
	int vy = y;

	if (!sys.fCGA)
		colEven = colOdd = EGACOLOR(color);

	PutPixel(vx, vy, colEven);
	PutPixel(vx + 1, vy, colOdd);
}

Uint32 GetPixel(int x, int y)
{
	return _GetPixel(sys.screen, x, y);
}

void LockScreen()
{
	_LockScreen(sys.screen);
}

void UnlockScreen()
{
	_UnlockScreen(sys.screen);
}

// Interface

bool GetDebug()
{
	return sys.fDebug;
}

void SetDebug(bool fDebug)
{
	sys.fDebug = fDebug;
}

void SetSound(bool fSound)
{
	sys.fSound = fSound;
}

bool GetSound()
{
	return sys.fSound;
}

void SetCGA(bool fCGA)
{
	sys.fCGA = fCGA;
}

bool GetCGA()
{
	return sys.fCGA;
}

void SetDblSize(bool fDblSize)
{
	sys.fDblSize = fDblSize;
}

void FlipSound()
{
	sys.fSound = !sys.fSound;
}

void FlipCGA()
{
	sys.fCGA = !sys.fCGA;
}

void FlipDblSize()
{
	sys.fDblSize = !sys.fDblSize;
	InitScreen();
}

// Initialization

void InitGraphics()
{
	sys.fCGA = false;
	sys.fDblSize = true;
	sys.fSound = true;
	sys.fDebug = false;
}
